/*
 * Project: irond
 * Package: src.de.fhhannover.inform.iron.mapserver.datamodel
 * File:    ClientService.java
 *
 * Copyright (C) 2010-2011 Fachhochschule Hannover
 * Ricklinger Stadtweg 118, 30459 Hannover, Germany 
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.fhhannover.inform.iron.mapserver.datamodel;

import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.Logger;

import de.fhhannover.inform.iron.mapserver.datamodel.graph.meta.Metadata;
import de.fhhannover.inform.iron.mapserver.exceptions.PurgePublisherNoAllowedException;
import de.fhhannover.inform.iron.mapserver.exceptions.ResponseCreationException;
import de.fhhannover.inform.iron.mapserver.logging.Logging;

/**
 * This class is responsible to handle newSession, endSession and purgePublisher.
 * DataModelService simply delegates the above calls to newSession(), endSession()
 * and purgePublish() respectively.
 * 
 * @author aw
 * @version 0.1
 */
class ClientService {
	
	/**
	 * static logger instance
	 */
	private static final Logger sLogger = Logging.getTheLogger();
	private static final String sName = "ClientService";

	private PublisherRep publisherRep;
	private MetaController metaController;
	private SubscriptionService subService;
	
	ClientService(PublisherRep pr, MetaController mCtl, SubscriptionService subServ) {
		publisherRep = pr;
		metaController = mCtl;
		subService = subServ;
	}
	
	/**
	 * Method which handles newSession calls.
	 * Simply a new publisher is added to the publisherRepository.
	 * In case of errors exceptions are thrown.
	 * 
	 * @throws RunningSessionException newSession is called, the publisher exists
	 * 			but had never received an end session call
	 */
	void newSession(String sessionId, String publisherId, Integer mprs) {
		if (mprs == null)
			mprs = new Integer(DataModelService.getServerConfiguration().getDefaultMaxPollResultSize());
		sLogger.trace(sName + ": newSession for " + publisherId 
				+ " and maxPollResultSize=" + mprs + " bytes");
	
		publisherRep.addPublisher(publisherId, sessionId, mprs);
	}
	 
	/**
	 * Method which handles an endSession call.
	 * First check whether the publisher is available, if not
	 * throw a NoSuchPublisherException.
	 * 
	 * Delete all session metadata from the publisher with the help
	 * of MetaController
	 * 
	 * Remove the session from the publisher and the corresponding
	 * session in the PublisherRep.
	 * 
	 * @param request
	 * @return
	 */
	void endSession(String sessionId) {
		List<Metadata> changes = new ArrayList<Metadata>();
		
		// throws exception if no publisher found
		Publisher p = publisherRep.getPublisherBySessionId(sessionId);
		sLogger.debug(sName + ": endSession for " + p.getPublisherId());

		// drop all subscriptions
		subService.removeSubscriptions(p);
		
		// delete session metadata of this publisher
		List<Metadata> toremove = p.getSessionMetadata();
		sLogger.trace(sName + ": Removing " + toremove.size()
				+ " session metadata objects");
		metaController.deleteMetadata(p, toremove, changes);
		
		// remove the session
		sLogger.trace(sName + ": Removing session from " + p.getPublisherId());
		p.deleteSessionId();
		publisherRep.removePublisherSession(sessionId);
		
		subService.commitChanges(changes);
	}

	/**
	 * Remove all metadata from publisher
	 * 
	 * @param request
	 * @return
	 * @throws PurgePublisherNoAllowedException 
	 * @throws ResponseCreationException 
	 */
	void purgePublisher(String sessionId, String publisherId) throws PurgePublisherNoAllowedException {
		List<Metadata> changes = new ArrayList<Metadata>();
		if (DataModelService.getServerConfiguration().getPurgePublisherIsRestricted()) {
			Publisher publisher = publisherRep.getPublisherBySessionId(sessionId);
			
			if (!publisher.getPublisherId().equals(publisherId)) {
				sLogger.warn(sName + ": purgePublisher not allowed");
				throw new PurgePublisherNoAllowedException("Not allowed to" +
						" purge metadata! of a different publisher");
			}
		}
		Publisher publisher = publisherRep.getPublisherByPublisherIdUnsafe(publisherId);
		
	
		// if we don't know about the publisher, stay silent, just do nothing
		if (publisher == null) {
			sLogger.warn(sName + ": PurgePublisher: Publisher with publisher-id=\""
					+ publisherId + "\" not found!");
			return;
		}
		sLogger.debug(sName + ": Purging metadata of " + publisher);

		
		List<Metadata> mdlist = publisher.getSessionMetadata();
		sLogger.trace("Deleting n=" + mdlist.size() + " session metadata objects");
		metaController.deleteMetadata(publisher, mdlist, changes);
		
		mdlist = publisher.getLifetimeMetadata();
		sLogger.trace("Deleting n=" + mdlist.size() + " lifetime metadata objects");
		metaController.deleteMetadata(publisher, mdlist, changes);
		
		subService.commitChanges(changes);
	}
}
