/*
 * Project: irond
 * Package: src.de.fhhannover.inform.iron.mapserver.datamodel
 * File:    PollResultImpl.java
 *
 * Copyright (C) 2010-2011 Fachhochschule Hannover
 * Ricklinger Stadtweg 118, 30459 Hannover, Germany 
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.fhhannover.inform.iron.mapserver.datamodel;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import de.fhhannover.inform.iron.mapserver.IfmapConstStrings;
import de.fhhannover.inform.iron.mapserver.messages.PollResult;
import de.fhhannover.inform.iron.mapserver.messages.SearchResult;
import de.fhhannover.inform.iron.mapserver.messages.SearchResultType;
import de.fhhannover.inform.iron.mapserver.utils.LengthCheck;
import de.fhhannover.inform.iron.mapserver.utils.NullCheck;

/**
 * Implementation to create {@link PollResult} objects with some logic to
 * add {@link SearchResult} objects and compress them while if possible.
 * <br/>
 * <b>Note: errorResult are not taken into account for the size calculation</b>
 * (TODO?)
 * 
 * @author aw
 *
 */
class PollResultImpl implements PollResult {
	
	private List<SearchResult> mResults;
	private List<String> mErrors;
	private int mCurSize;

	PollResultImpl() {
		mResults = new ArrayList<SearchResult>();
		mErrors = new ArrayList<String>();
		mCurSize = 0;
	}
	
	/**
	 * @return a read-only list for all contained {@link SearchResult} instances
	 */
	public List<SearchResult> getResults() {
		return Collections.unmodifiableList(mResults);
	}

	/**
	 * @return a read-only list for all contained errors
	 */
	public List<String> getErrorResults() {
		return Collections.unmodifiableList(mErrors);
	}
	
	void addErrorResult(String name) {
		NullCheck.check(name, "name is null");
		LengthCheck.checkMinMax(name, 1, 20, "bad sub name for error result");
		
		// Sanity
		for (SearchResult res : mResults)
			if (res.getName().equals(name))
				throw new RuntimeException("SearchResults and ErrorResult in "
						+ "PollResult for subscription " + name);

		mErrors.add(name);
	}
	
	/**
	 * Add a new {@link SearchResult} to the {@link PollResult}.
	 * If the last {@link SearchResult} is of the same {@link SearchResultType}
	 * and the same {@link Subscription} then the {@link SearchResult} can be
	 * compressed into the last one.
	 * 
	 * @param sres
	 */
	void addSearchResult(SearchResult sres) {
		NullCheck.check(sres, "sres is null");
		NullCheck.check(sres.getName(), "sres.getName() is null");
		LengthCheck.checkMinMax(sres.getName(), 1, 20, "sub name bad");
		
		SearchResult lastResult = null;
		
		if (mResults.size() > 0)
			lastResult = mResults.get(mResults.size() - 1);
		
		if (lastResult != null && sameAsLastResult(sres, lastResult)) {
			mCurSize -= lastResult.getByteCount();
			lastResult.addResultItems(sres.getResultItems());
		} else {
			mResults.add(sres);
			lastResult = sres;
		}
		mCurSize += lastResult.getByteCount();
	}

	private boolean sameAsLastResult(SearchResult newRes, SearchResult lastRes) {
		return (lastRes != null && lastRes.getType() == newRes.getType()
				&& lastRes.getName().equals(newRes.getName()));
	}

	public void removeResults(String subName) {
		List<SearchResult> toRemoveSres = new ArrayList<SearchResult>();
		List<String> toRemoveErrs = new ArrayList<String>();
		for (SearchResult sr : mResults)
			if (sr.getName().equals(subName)) {
				toRemoveSres.add(sr);
				mCurSize -= sr.getByteCount();
			}
		
		mResults.removeAll(toRemoveSres);
		
		for (String error : mErrors)
			if (error.equals(subName)) {
					toRemoveErrs.add(error);
			}
		
		mErrors.removeAll(toRemoveErrs);
	}

	public Integer getByteCount() {
		return IfmapConstStrings.PRES_MIN_CNT + mCurSize;
	}

	public boolean isEmpty() {
		return mResults.isEmpty() && mErrors.isEmpty();
	}

	/**
	 * Returns the bytes used for all {@link SearchResult} objects in this
	 * pollResult.
	 * 
	 * @param name
	 * @return
	 */
	public int getByteCountFor(String name) {
		int sum = 0;
		for (SearchResult sr : mResults)
			if (sr.getName().equals(name))
				sum += sr.getByteCount();
		return sum;
	}
}
