/*
 * Project: irond
 * Package: src.de.fhhannover.inform.iron.mapserver.datamodel
 * File:    PublishService.java
 *
 * Copyright (C) 2010-2011 Fachhochschule Hannover
 * Ricklinger Stadtweg 118, 30459 Hannover, Germany 
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.fhhannover.inform.iron.mapserver.datamodel;

import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import org.apache.log4j.Logger;

import de.fhhannover.inform.iron.mapserver.datamodel.graph.Identifier;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.MetadataContainer;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.meta.MetaState;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.meta.Metadata;
import de.fhhannover.inform.iron.mapserver.exceptions.DoRollbackException;
import de.fhhannover.inform.iron.mapserver.exceptions.InvalidMetadataException;
import de.fhhannover.inform.iron.mapserver.exceptions.ResponseCreationException;
import de.fhhannover.inform.iron.mapserver.logging.Logging;
import de.fhhannover.inform.iron.mapserver.messages.PublishDelete;
import de.fhhannover.inform.iron.mapserver.messages.PublishNotify;
import de.fhhannover.inform.iron.mapserver.messages.PublishRequest;
import de.fhhannover.inform.iron.mapserver.messages.PublishUpdate;
import de.fhhannover.inform.iron.mapserver.messages.SubPublishRequest;

/**
 * Publish or delete metadata.
 *
 *
 * TODO: Check what should be done with replaced singleValue metadata.
 *       Currently they are silently removed. Do they have to appear in
 *       poll results of MAPCs?
 * 
 * @author awe, vp
 * @version 0.1
 * created: 19.11.09
 */
class PublishService {
	
	/**
	 * a static logger instance
	 */
	private final static Logger sLogger = Logging.getTheLogger();
	private final static String sName = "PublishService";
	
	private PublisherRep mPublisherRep;
	private IdentifierRep mIdentifierRep;
	private MetaController mMetaController;
	private LinkRep mLinkRep;
	private SubscriptionService mSubService;
	
	PublishService(PublisherRep pRep, IdentifierRep iRep, LinkRep lRep,
			MetaController mCtl, SubscriptionService subServ) {
		mPublisherRep = pRep;
		mIdentifierRep = iRep;
		mLinkRep = lRep;
		mMetaController	= mCtl;
		mSubService	= subServ;
	}
	
	/**
	 * Go through the whole PublishRequest and dispatch the elements to
	 * the corresponding process* methods
	 *
	 * @param request
	 * @return PublishReceived or PublishError
	 * @throws NotSupportedException 
	 * @throws InvalidMetadataException 
	 * @throws ResponseCreationException 
	 */
	void publish(PublishRequest request) throws InvalidMetadataException {
		sLogger.trace(sName + ": PublishRequest from sessionid=" + request.getSessionId());
		List<SubPublishRequest> list = request.getSubPublishRequestList();
		
		Publisher publisher = 
			mPublisherRep.getPublisherBySessionId(request.getSessionId());
		
		sLogger.trace(sName + ": Traversing n=" + request.getSubPublishRequestList().size() + 
				" PublishRequests for " + publisher.getPublisherId());
	
		List<Metadata> changes = new ArrayList<Metadata>();
		
		try {
			for (SubPublishRequest req : list) {
				
				if (req instanceof PublishNotify)
					processPublishNotify(publisher, (PublishNotify) req, changes);
				else if (req instanceof PublishUpdate)
					processPublishUpdate(publisher, (PublishUpdate)req, changes);
				else if (req instanceof PublishDelete)
					processPublishDelete(publisher, (PublishDelete)req, changes);
				else
					throw new RuntimeException("Unknown SubPublishRequest implementation");
			}
		
			// set timestamp and publisher id _now_;
			String timeNow = getISO8601DateTime(new Date());
			for (Metadata m : changes) {
				m.setTimestamp(timeNow);
				m.setPublisherId(publisher.getPublisherId());
			}
			
			mIdentifierRep.setTimestamp(System.currentTimeMillis());
			mSubService.commitChanges(changes);
		} catch (DoRollbackException e) {
			sLogger.debug(sName + ": Doing Rollback!");
			mSubService.rollbackChanges();
			throw new InvalidMetadataException(e.getMessage());
		}
	}
	
	
	/**
	 * Publish metadata to identifier or link.
	 * 
	 * - get a MetadataContainer with help of getMetadataContainer()
	 *   (whether link or identifier does not matter)
	 * 
	 * - go through the list of metadata, modify every metadata object
	 *   to include the publisherId, a timestamp and set the lifetime
	 *   thereafter call addMetadata() on the target container.
	 *   a copy of the metadata will be created an returned.
	 *   
	 * - add the new metadata object to the publisher who published it
	 * 
	 *
	 * @param request
	 * @throws DoRollbackException 
	 */
	private void processPublishUpdate(Publisher pub, PublishUpdate request,
			List<Metadata> changes) throws DoRollbackException {
		
		List<Metadata> mlist = request.getMetadataList();
		
	
		// set the state of all metadata objects, if it's not NOTIFY (set by
		// processPublishNotify(), then it must be NEW
		for (Metadata m : mlist) {
			if (!m.isNotify()) {
				m.setState(MetaState.NEW);
			}
		}
		
		MetadataContainer mc = 
			getMetadataContainer(request.getIdent1(),request.getIdent2());
		
		

		for (Metadata m : mlist) {
			sLogger.trace(sName + ": Adding " + m.getPrefixAndElement() + " to " + mc);
			List<Metadata> removed = new ArrayList<Metadata>();
			
			Metadata added = mc.addMetadata(m, removed);
		
			if (added != null) {
				changes.add(added);
				added.setContainerReference(mc);
				added.setPublisherReference(pub);
				pub.addMetadata(added);
				
				// snaity check
				if (added.isNotify()) {
					if (removed.size() > 0)
						throw new RuntimeException("Notify metadata triggered"
								+ " removal of metadata in the graph");
				}
			
				// check whether new metadata was replaced, if yes remove it from
				// changes
				for (Metadata rmmd : removed) {
					if (rmmd.isNew()) {
						// Metadata with state NEW is simply dropped from the
						// changes list and from the associated container
						// sanity check, whether we really removed something
						if (!changes.remove(rmmd)) {
							throw new RuntimeException("Removed NEW Metadata was" +
									"not in changes. Programming Error?");
						}
						
						mMetaController.removeReferencesTo(rmmd);
					} else if (rmmd.isDeleted()){
						// leave deleted in the changes and "in the graph"
						// FIXME: is this ok?
					} else {
						// silently drop removed metadata
						mMetaController.removeReferencesTo(rmmd);
					}
				}
			} else {
				sLogger.error(sName + ": Metadata adding failed!");
			}
		}
	}
	 
	/**
	 * Delete metadata from link or identifier
	 * 
	 * @param reqPublisher the publisher responsible for this request
	 * @param request the delete request
	 * @param changes 
	 * @return
	 */
	private void processPublishDelete(Publisher reqPublisher, PublishDelete request,
			List<Metadata> changes) {
		
		List<Metadata> toremove;
		MetadataContainer metadatacontainer;
		
		metadatacontainer = getMetadataContainer(request.getIdent1(),
				request.getIdent2());
		
		// request metadata to be removed
		toremove = metadatacontainer.getMetadata(request.getFilter());
		
		
		sLogger.trace(sName + ": Deleting n=" + toremove.size() + " Metadata objects from "
				+ metadatacontainer);
	
		mMetaController.deleteMetadata(reqPublisher, toremove, changes);		
	}
	
	
	
	
	/**
	 * Get the list of {@link Metadata} objects, set their state to
	 * NOTIFY and forward the request to the processPublishUpdate
	 * method. There are some hacks over there that this process
	 * should work...
	 * 
	 * @param pub
	 * @param request
	 */
	private void processPublishNotify(Publisher pub, PublishNotify request,
			List<Metadata> changes) {
		
		List<Metadata> mlist = request.getMetadataList();
		
		
		for (Metadata m : mlist) {
			m.setState(MetaState.NOTIFY);
		}
		try {
			processPublishUpdate(pub, request, changes);
		} catch (DoRollbackException e) {
			// nah, that should not occur...
			sLogger.error("UNEXPECTED: Rollback triggered while using notify metadata");
			e.printStackTrace();
		}
	}
	
	

	/**
	 * Check data type of SubPublishRequest and return MetadataContainer
	 * 
	 * @param ident1
	 * @param ident2
	 * @return MetadataContainer
	 */
	private MetadataContainer getMetadataContainer(Identifier ident1, Identifier ident2) {
		MetadataContainer metadatacontainer = null;
		if (ident1 != null && ident2 != null) {
			metadatacontainer = 
				mLinkRep.getOrCreateLink(ident1, ident2);
		} else if (ident1 != null || ident2 != null) {
			
			if (ident1 != null) {
				metadatacontainer = 
					mIdentifierRep.getOrCreateIdentifier(ident1);
			} else {
				metadatacontainer = 
					mIdentifierRep.getOrCreateIdentifier(ident2);
			}
		} else {
			metadatacontainer = null;
		}
		return metadatacontainer;
	}
	
	/**
	 * Thanks Ingo!
	 * @param dt
	 * @return
	 */
	private String getISO8601DateTime(Date dt){
		//FIXME hack in order to get good timestamps for xsd:dateTime
        SimpleDateFormat formatter = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss");
        SimpleDateFormat tzFormatter = new SimpleDateFormat("Z");
        String timezone = tzFormatter.format(dt);
        return formatter.format(dt) + timezone.substring(0, 3) + ":"
                + timezone.substring(3);
	}
}
 

