/*
 * Project: irond
 * Package: src.de.fhhannover.inform.iron.mapserver.datamodel
 * File:    Subscription.java
 *
 * Copyright (C) 2010-2011 Fachhochschule Hannover
 * Ricklinger Stadtweg 118, 30459 Hannover, Germany 
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.fhhannover.inform.iron.mapserver.datamodel;

import java.util.ArrayList;
import java.util.List;

import de.fhhannover.inform.iron.mapserver.datamodel.graph.MetadataContainer;
import de.fhhannover.inform.iron.mapserver.messages.SearchRequest;
import de.fhhannover.inform.iron.mapserver.messages.SearchResult;
import de.fhhannover.inform.iron.mapserver.utils.LengthCheck;
import de.fhhannover.inform.iron.mapserver.utils.LinkedMap;
import de.fhhannover.inform.iron.mapserver.utils.NullCheck;
/**
 * A subscription object encapsulates a {@link SearchRequest} and 
 * four different {@link SearchRequest} objects. Update, delete, notify and
 * search.
 * 
 * 
 * @version 0.1
 * @author aw, vp
 */

/*
 * created: 03.05.10
 * 
 * changes:
 *  03.05.10 aw, vp  ...
 *  08.06.10 aw - some commenting and minor changes
 *  11.06.10 aw - add names to the search results :-/
 * 
 * 
 */
public class Subscription {
	
	/**
	 * Name of the subscription
	 */
	private final String mName;
	
	/**
	 * Corresponding searchRequest
	 */
	private final SearchRequest mSearchRequest;

	/**
	 * {@link LinkedMap} of {@link MetadataContainer} for this subscription
	 * contains.
	 */
	private final LinkedMap<MetadataContainer> mContainers;
	
	private boolean changed;
	
	private boolean exceededSize;
	
	/**
	 * Reference to the publisher who is the owner of this subscription.
	 */
	private final Publisher mPublisherReference;
	
	private SearchResult reRunOldResult;
	private SearchResult reRunNewResult;

	Subscription(Publisher publisher, String name, SearchRequest searchRequest) {
		NullCheck.check(publisher, "publisher is null");
		NullCheck.check(name, "name is null");
		NullCheck.check(searchRequest, "searchRequest is null");
		LengthCheck.checkMinMax(name, 1, 20, "subscription name length bad");
		
		mName = name;
		mPublisherReference = publisher;
		mSearchRequest = searchRequest;
		mContainers = new LinkedMap<MetadataContainer>();
		
		changed = false;
		exceededSize = false;
	}
			

	/**
	 * @return the name
	 */
	String getName() {
		return mName;
	}

	/*
	 * auto generated getters and setters
	 */
	SearchRequest getSearchRequest() {
		return mSearchRequest;
	}
	
	List<MetadataContainer> getContainers() {
		return new ArrayList<MetadataContainer>(mContainers.values());
	}
	
	void addContainer(MetadataContainer mc) {
		NullCheck.check(mc, "mc is null");
		if (mContainers.contains(mc))
			throw new RuntimeException("Container " + mc + " already there");
		
		mContainers.put(mc);
	}

	void removeContainer(MetadataContainer mc) {
		NullCheck.check(mc, "mc is null");
		if (!mContainers.contains(mc))
			throw new RuntimeException("Cannot remove non existent container");
		
		mContainers.remove(mc);
	}

	void setReRunOldResult(SearchResult reRunOldResult) {
		this.reRunOldResult = reRunOldResult;
	}

	SearchResult getReRunOldResult() {
		return reRunOldResult;
	}

	void setReRunNewResult(SearchResult reRunNewResult) {
		this.reRunNewResult = reRunNewResult;
	}

	SearchResult getReRunNewResult() {
		return reRunNewResult;
	}

	void setChanged() {
		changed = true;
	}
	
	boolean hasChanged() {
		return changed;
	}

	Publisher getPublisherReference() {
		return mPublisherReference;
	}
	
	Integer getMaxResultSize() {
		return mSearchRequest.getMaxResultSize();
	}
	
	boolean exceededSize() {
		return exceededSize;
	}
	
	void setExceededSize() {
		exceededSize = true;
	}
}
