/*
 * Project: irond
 * Package: src.de.fhhannover.inform.iron.mapserver.datamodel
 * File:    SubscriptionService.java
 *
 * Copyright (C) 2010-2011 Fachhochschule Hannover
 * Ricklinger Stadtweg 118, 30459 Hannover, Germany 
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.fhhannover.inform.iron.mapserver.datamodel;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import org.apache.log4j.Logger;

import de.fhhannover.inform.iron.mapserver.datamodel.graph.Identifier;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.Link;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.MetadataContainer;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.meta.MetaState;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.meta.Metadata;
import de.fhhannover.inform.iron.mapserver.exceptions.AlreadyObservedException;
import de.fhhannover.inform.iron.mapserver.exceptions.NoSuchSubscriptionException;
import de.fhhannover.inform.iron.mapserver.exceptions.PollResultsTooBigException;
import de.fhhannover.inform.iron.mapserver.exceptions.ResponseCreationException;
import de.fhhannover.inform.iron.mapserver.exceptions.SearchResultsTooBigException;
import de.fhhannover.inform.iron.mapserver.logging.Logging;
import de.fhhannover.inform.iron.mapserver.messages.PollResult;
import de.fhhannover.inform.iron.mapserver.messages.ResultItem;
import de.fhhannover.inform.iron.mapserver.messages.SearchRequest;
import de.fhhannover.inform.iron.mapserver.messages.SearchResult;
import de.fhhannover.inform.iron.mapserver.messages.SearchResultType;
import de.fhhannover.inform.iron.mapserver.messages.SubPublishRequest;
import de.fhhannover.inform.iron.mapserver.messages.SubSubscribeRequest;
import de.fhhannover.inform.iron.mapserver.messages.SubscribeDelete;
import de.fhhannover.inform.iron.mapserver.messages.SubscribeRequest;
import de.fhhannover.inform.iron.mapserver.messages.SubscribeUpdate;
import de.fhhannover.inform.iron.mapserver.utils.NullCheck;

/**
 * Class to handle SubscriptionRequests
 * 
 * @version 0.1
 * @author aw
 *
 */
public class SubscriptionService {
	
	private static Logger sLogger = Logging.getTheLogger();
	private static String sName = "SubscriptionService";
	
	private final PublisherRep publisherRep;
	private final MetaController metaController;
	private final IdentifierRep identifierRep;
	
	private SubscriptionObserver mObserver;
	
	private final Set<Subscription> changedSubscriptions;
	private final Set<Publisher> changedPublishers;
	private List<Metadata> mChangedMetadata;
	private List<Subscription> mRerunSubs;

	SubscriptionService(IdentifierRep iRep, PublisherRep pRep, MetaController mCtl) {
		identifierRep = iRep;
		publisherRep = pRep;
		metaController = mCtl;
		
		changedSubscriptions = new HashSet<Subscription>();
		changedPublishers = new HashSet<Publisher>();
		mChangedMetadata = new ArrayList<Metadata>();
		mRerunSubs = new ArrayList<Subscription>();
	}
	
	/**
	 * Handling of a {@link SubscriptionRequest} which is forwarded by the
	 * {@link DataModelService}.
	 * Look at all teh {@link SubPublishRequest} objects contained in the
	 * {@link SubscriptionRequest} and dispatch whether it's a update or
	 * a delete.
	 * 
	 * @param subscriptionReq
	 * @throws NoSuchSubscribeException
	 */
	void subscribe(SubscribeRequest subReq) throws NoSuchSubscriptionException {
		NullCheck.check(subReq, "subReq is null");
		
		Publisher pub = publisherRep.getPublisherBySessionId(subReq.getSessionId());
		sLogger.debug(sName + ": subscribe for " + pub.getPublisherId());
		
		for (SubSubscribeRequest ssr : subReq.getSubSubscribeRequests()) {
			if (ssr instanceof SubscribeUpdate)
					processSubscribeUpdate(pub, (SubscribeUpdate)ssr);
			else if (ssr instanceof SubscribeDelete)
					processSubscribeDelete(pub, (SubscribeDelete)ssr);
			else
				throw new RuntimeException("Unknown SubscribeRequest Impl");
		}
	}

	/**
	 * Delete a {@link Subscription} from a {@link Publisher}
	 * 
	 * @param pub
	 * @param ssr
	 * @throws NoSuchSubscriptionException 
	 */
	private void processSubscribeDelete(Publisher pub, SubscribeDelete ssr)
											throws NoSuchSubscriptionException {
		
		NullCheck.check(pub, "pub is null");
		NullCheck.check(ssr, "ssr is null");
		
		SubscriptionState subState = pub.getSubscriptionState();
		String name = ssr.getName();
		Subscription sub = subState.getSubscription(name);
		
		if (sub != null) {
			// Sanity if mapping worked before
			if (!sub.getName().equals(ssr.getName()))
				throw new RuntimeException("Bad subscription mapping for "
						+ ssr.getName() + ", got " + sub.getName());
				
			sLogger.debug(sName + ": Deleting subscription \"" + sub.getName()
					+ "\" of " + pub);
			clearContainers(sub);
			subState.removeSubscription(sub);
		} else {
			throw new NoSuchSubscriptionException("Subscription " +
					ssr.getName() + " does not exist");
		}
	}

	/**
	 * Create a new {@link Subscription}, attach it to a {@link Publisher},
	 * run the first search.
	 * 
	 * If a {@link Subscription} with the same name is already there, reset
	 * it and update the {@link SearchResult} objects
	 * 
	 * @param pub
	 * @param ssr
	 */
	private void processSubscribeUpdate(Publisher pub, SubscribeUpdate ssr) {

		NullCheck.check(pub, "pub is null");
		NullCheck.check(ssr, "ssr is null");
		NullCheck.check(ssr.getSearchRequest(), "sreq is null");
		
		SubscriptionState subState = pub.getSubscriptionState();
		Subscription sub = subState.getSubscription(ssr.getName());
		
		sLogger.debug(sName + ": subscribe update \"" + ssr.getName() + "\" for "
				+ pub.getPublisherId());
		
		// TODO: Maybe we should throw something here to notify the MAPC about
		// this problem.
		if (subState.isPollResultsTooBig()) {
			sLogger.warn(sName + ": Ignoring subscribeUpdate as the PollResult "
					+ " grew too big and no notification was sent out yet");
			return;
		}
		
		// Check if subscription was already there and remove it if it was.
		if (sub != null) {
			sLogger.debug(sName + ": subscribe update for existing subscription"
					+ "\"" + sub.getName() + "\", removing existing subscription");
			clearContainers(sub);
			subState.removeSubscription(sub);
		}
			
		sub = new Subscription(pub, ssr.getName(), ssr.getSearchRequest());
		subState.addSubscription(sub);
	
		try {
			SearchResult initResult = runInitialSearch(sub);
			subState.getPollResult().addSearchResult(initResult);
		} catch (SearchResultsTooBigException e) {
			sub.setExceededSize();
			subState.getPollResult().addErrorResult(sub.getName());
			sLogger.debug(sName + ": initial search too big: " + e.getReached()
					+ " bytes used, but only " + e.getLimit() + " bytes allowed.");
		}
			
		sub.setChanged();
		setHasChanges(sub);
		doNotifications();
		cleanup();
	}
	
	/**
	 * If a new {@link Subscription} is created, the first thing to do is to
	 * run a search in the current graph. While doing so, mark all contained
	 * {@link MetadataContainer} with this {@link Subscription}.
	 * Further, put the first {@link SearchResult} of type
	 * {@link SearchResultType#SEARCH} into the {@link PollResult}.
	 * 
	 * @param sub
	 */
	private SearchResult runInitialSearch(Subscription sub) throws SearchResultsTooBigException {
		
		SearchRequest sreq = sub.getSearchRequest();
	
		SearchResult initSres = new SearchResultImpl(sub.getName());
		SearchResult refSres = new SearchResultReferenceImpl(sub.getName());
		
		SearchResultBuilder initBuilder = new SearchResultBuilderImpl(initSres);
		SearchResultBuilder refBuilder = new SearchResultBuilderImpl(refSres);
		Searcher searcher = new Searcher(identifierRep, sreq, refBuilder);
		
		// Run a search to find references to the MetadataContainer objects
		searcher.runSearch();
			
		for (ResultItem ri : refSres.getResultItems()) {
			MetadataContainer mc = ri.getContainer();
			initBuilder.addContainer(mc);
			initBuilder.addMetadata(mc, ri.getMetdata());
		}
		initBuilder.filter(sreq.getResultFilter());
		
		registerContainers(sub, refBuilder.getResult());
		return initBuilder.getResult();
	}

	/**
	 * TODO: for metadata to appear in a notifyResult the state of the
	 *       subscriptions before <b>all</b> changes because of publishUpdate
	 *       or publishDelete is taken. It's been like this some time before,
	 *       but one could think of the following scenario:
	 *       - A publishUpdate introduces a link somewhere, such that the
	 *         subscription would be active for more identifiers and links.
	 *       - A consecutive notify of metadata would appear in the notifyResult
	 *         because of the "new state" of the subscription.
	 *       As mentioned above, currently only the old state of subscriptions
	 *       is used.
	 *       
	 * @param changedMetadata
	 * @throws ResponseCreationException 
	 */
	void commitChanges(List<Metadata> changes) {
		sLogger.debug(sName + ": Commit Changes...");
		mChangedMetadata = changes;
		
		dumpChangedMetadata();
		reRunFindSubscriptions();
		reRunOldStateSearch();
		doSimpleUpdates();
		updateGraphState();
		reRunNewStateSearch();
		reRunUpdateResults();
		markSubscriptionsGrewTooBig();
		markPublisherPollResultTooBig();
		doNotifications();
		cleanup();
	}

	private void updateGraphState() {
		for (Metadata m : mChangedMetadata) {
			switch (m.getState()) {
				case NEW:
					m.setState(MetaState.UNCHANGED);
					break;
				case DELETED:
					removeObsoleteSubscriptions(m);
					// fall through, as it is the same
				case NOTIFY:
					metaController.removeReferencesTo(m);
					break;
				default:
					throw new RuntimeException("Metadata with bad state in changes");
			}
		}
	}

	/**
	 * If metadata was deleted from a link and the link was a leaf edge in the
	 * graph, it might be necessary to remove a subscription which has no matching
	 * metadata anymore on this link.
	 * 
	 * @param m
	 */
	private void removeObsoleteSubscriptions(Metadata m) {
		MetadataContainer mc = m.getContainerReference();
		List<Subscription> subs = mc.getSubscriptions();
		
		if (!(mc instanceof Link))
			return;
		
		for (Subscription sub : subs) {
			int count = 0;
			Filter lFilter = sub.getSearchRequest().getMatchLinksFilter();
			List<Metadata> remainingMetadata = mc.getMetadata(lFilter);
			
			for (Metadata rMd : remainingMetadata)
				if (!rMd.isDeleted() && !rMd.isNotify())
					count++;
				
			if (count > 0) {
				mc.removeSubscription(sub);
				sub.removeContainer(mc);
			}
		}
	}

	private void dumpChangedMetadata() {
		if (sLogger.isTraceEnabled()) {
			sLogger.trace(sName + ": Changed metadata:");
			for (Metadata m : mChangedMetadata) {
				sLogger.trace("\t" + m.getPrefixAndElement() 
						+ " [state=" + m.getState()
						+ " cardinality=" + m.getCardinality()
						+ " lifetime=" + m.getLifeTime()
						+ " size=" + m.getByteCount() + " bytes]");
			}
		}
	}

	private void doSimpleUpdates() {
		for (Metadata m : mChangedMetadata)
			addToMatchingSubscriptions(m);
	}

	private void addToMatchingSubscriptions(Metadata m) {
		MetadataContainer mc = m.getContainerReference();
		for (Subscription sub : mc.getSubscriptions())
			if (matchesOnSubscription(m, sub))
					addToPollResult(sub, m);
	}
	
	private boolean matchesOnSubscription(Metadata m, Subscription sub) {
		MetadataContainer mc = m.getContainerReference();
		Filter lFilter = sub.getSearchRequest().getMatchLinksFilter();
		Filter rFilter = sub.getSearchRequest().getResultFilter();
		
		if (mc instanceof Identifier)
			return m.matchesFilter(rFilter);
		else if (mc instanceof Link)
			return (m.matchesFilter(lFilter) && m.matchesFilter(rFilter));
		else
			throw new RuntimeException("MetadataContainer not Link nor Identifier");
	}

	private void addToPollResult(Subscription sub, Metadata m) {
		SubscriptionState subState = sub.getPublisherReference().getSubscriptionState();
		SearchResultType type = null;
		PollResultImpl pollResult = subState.getPollResult();
		SearchResult toAdd = null;
		
		switch (m.getState()) {
			case NEW:
				type = SearchResultType.UPDATE;
				break;
			case DELETED:
				type = SearchResultType.DELETE;
				break;
			case NOTIFY:
				type = SearchResultType.NOTIFY;
				break;
			default:
				throw new RuntimeException("Bad Metadata state in changed Metadata");
		}
		
		toAdd = new SearchResultImpl(sub.getName(), type);
		toAdd.addMetadata(m.getContainerReference(), m);
		pollResult.addSearchResult(toAdd);
		sub.setChanged();
		setHasChanges(sub);
	}

	/**
	 * For {@link Subscription} instances that have a new sub graph added
	 * or removed a new search has to be done. Then the {@link Subscription}
	 * instances are removed from the {@link MetadataContainer} objects.
	 */
	private void reRunOldStateSearch() {
		for (Subscription sub : mRerunSubs) {
			sLogger.trace(sName + ": doing old state search for subscription \""
					+ sub.getName() + "\"");
			SearchResult oldRes = doReferenceSearch(sub);
			sub.setReRunOldResult(oldRes);
		}
	}
	
	private void reRunNewStateSearch() {
		for (Subscription sub : mRerunSubs) {
			sLogger.trace(sName + ": doing new state search for subscription \""
					+ sub.getName() + "\"");
			SearchResult newRes = doReferenceSearch(sub);
			sub.setReRunNewResult(newRes);
			registerContainers(sub, newRes);
		}
	}
	
	/**
	 * Helper to do a reference search without a size limit.
	 * 
	 * @param sub
	 * @return
	 */
	private SearchResult doReferenceSearch(Subscription sub) {
		SearchRequest sReq = sub.getSearchRequest();
		SearchResult result = new SearchResultReferenceImpl();
		SearchResultBuilder builder = new SearchResultBuilderImpl(result);
		Searcher searcher = new Searcher(identifierRep, sReq, builder);
		searcher.runSearchIgnoreSize();
		return result;
	}
	
	/**
	 * Helper method which compares the Old and New SearchResult and appends
	 * the differences into the Delete and Update Result.
	 * 
	 * {@link Subscription} objects contain new and old {@link SearchResult}
	 * objects which are set if the {@link Subscription} is rerun.
	 * Those are of the type {@link SearchResultImplReferences}, so we can
	 * compare with ==, as we compare references.
	 * 
	 * @param sub
	 */
	private void reRunUpdateResults() {
		
		for (Subscription sub : mRerunSubs) {
			SubscriptionState subState = sub.getPublisherReference().getSubscriptionState();
			SearchResult newRes = sub.getReRunNewResult();
			SearchResult oldRes = sub.getReRunOldResult();
			SearchResult diff[] = calculateDiff(sub.getName(), newRes, oldRes);
		
			if (!diff[0].isEmpty())
				subState.getPollResult().addSearchResult(diff[0]);
			if (!diff[1].isEmpty())
				subState.getPollResult().addSearchResult(diff[1]);
			
			if (!diff[0].isEmpty() || !diff[1].isEmpty())
				setHasChanges(sub);
		}
	}
		
	/**
	 * URGS :(
	 * 
	 * This is still from hell...
	 * 
	 * @param subName
	 * @param newRes
	 * @param oldRes
	 * @return
	 */
	private SearchResult[] calculateDiff(String subName, SearchResult newRes,
			SearchResult oldRes) {
		List<ResultItem> oldItems = oldRes.getResultItems();
		List<ResultItem> newItems = new ArrayList<ResultItem>(newRes.getResultItems());
		
		SearchResult deleteResult = new SearchResultImpl(subName, SearchResultType.DELETE);
		SearchResult updateResult = new SearchResultImpl(subName, SearchResultType.UPDATE);
		SearchResult ret[] = new SearchResult[2];
		ret[0] = deleteResult;
		ret[1] = updateResult;
		
		for (ResultItem oldRi : oldItems) {
			ResultItem removeFromNew = null;
			for (ResultItem newRi : newItems) {
				MetadataContainer oldMc = oldRi.getContainer();
				MetadataContainer newMc = newRi.getContainer();
				
				if (oldMc.equals(newMc)) {
					List<Metadata> oldMetaList = oldRi.getMetdata();
					List<Metadata> newMetaList = new ArrayList<Metadata>(newRi.getMetdata());
					
					// compare metadata
					for (Metadata oldMeta : oldMetaList) {
						
						// oldMeta was contained in newMeta, forget about it.
						// otherwise it was deleted, so add it to the deleteResult
						if (newMetaList.contains(oldMeta))
							newMetaList.remove(oldMeta);
						else
							deleteResult.addMetadata(oldMc, oldMeta);
					}
					// Those that are left in newMetaList are new, put them
					// into an updateResult, but only if there is metadata
					// left to be added, otherwise a new resultItem may
					// be created.
					if (newMetaList.size() > 0)
						updateResult.addMetadata(oldMc, newMetaList);
					
					removeFromNew = newRi;
					
					break;
				} else {
					continue;
				}
			}
			
			if (removeFromNew != null) {
				newItems.remove(removeFromNew);
			} else {
				// The new list didn't contain oldRi, put it into the deleteResult,
				// but only if it contains metadata.
				if (oldRi.getMetdata().size() > 0)
					deleteResult.addResultItem(oldRi);
			}
		}
		
		// All ResultItems left in newItems are new, so add them to the update
		// Result
		updateResult.addResultItems(newItems);
		
		return ret;
	}

	private void setHasChanges(Subscription sub) {
		NullCheck.check(sub, "sub is null");
		if (!changedSubscriptions.contains(sub)) {
			sLogger.trace(sName + ": Adding " + sub.getName() + " to changed subscriptions");
			changedSubscriptions.add(sub);
			changedPublishers.add(sub.getPublisherReference());
		}
	}

	/**
	 * Go through the list of {@link Metadata} objects that are NEW or DELETED,
	 * if such a objects is attached to a {@link Link} some checks have to be
	 * done to find out if the {@link Subscription} needs to be rerun.
	 * 
	 * If the {@link Metadata} object is new, and the {@link Identifier} objects
	 * of the link contain {@link Subscription} objects that the {@link Link} does
	 * not contain, and further the newly created {@link Metadata} object
	 * matches the matchlinks filter of such a {@link Subscription} the 
	 * {@link Subscription} needs to be rerun.
	 * 
	 * If a {@link Metadata} object was deleted, we need to check for every
	 * {@link Subscription} object of this {@link Link} whether another
	 * {@link Metadata} object of that {@link Link} matches the matchlinks filter
	 * of the current {@link Subscription}. If this is not the case the
	 * {@link Subscription} needs to be rerun as a sub grapo might have been removed
	 * 
	 * @param changedMetadata the {@link Metadata} objects that have been marked as changed
	 * @return
	 */
	private void reRunFindSubscriptions() {
		
		HashSet<Subscription> ret = new HashSet<Subscription>();
		
		for (Metadata m : mChangedMetadata) {
			if (m.getContainerReference() instanceof Link) {
				Link link = (Link) m.getContainerReference();
				if (m.isNew())
					ret.addAll(findSubsWithAddedSubGraph(link, m));
				else if (m.isDeleted())
					ret.addAll(findSubsWithRemovedSubGraph(link, m));
			}
		}
		mRerunSubs = new LinkedList<Subscription>(ret);
	}
	
	/**
	 * TODO: Documentation
	 * 
	 * @param link
	 * @param m
	 * @return
	 */
	private List<Subscription> findSubsWithRemovedSubGraph(Link link, Metadata m) {
		List<Subscription> ret = new ArrayList<Subscription>();
		List<Subscription> subs = link.getSubscriptions();
			
		for (Subscription sub : subs) {
			Filter lFilter = sub.getSearchRequest().getMatchLinksFilter();
			List<Metadata> all = link.getMetadata();
			int count = 0;
			
			for (Metadata md : all)
				if (!md.isInGraph() && !md.isNotify() && md.matchesFilter(lFilter))
						count++;

			// Zero if there is no Metadata left on this link for the
			// given subscription, in which case a sub-graph was removed,
			// but only if both identifiers contain the subscription, as
			// otherwise it's a leaf edge and we can ignore it.
			if (count == 0
				&& link.getIdent1().getSubscriptions().contains(sub)
				&& link.getIdent2().getSubscriptions().contains(sub))
				ret.add(sub);
		}
	
		// Subscriptions that need to be rerun can be removed from the graph
		for (Subscription sub : ret)
			clearContainers(sub);
		
		return ret;
	}
	
	/**
	 * TODO: Documentation:
	 * 
	 * @param link
	 * @param m
	 * @return
	 */
	private List<Subscription> findSubsWithAddedSubGraph(Link link, Metadata m) {
		List<Subscription> ret = new ArrayList<Subscription>();
		HashSet<Subscription> notOnLinkSet = new HashSet<Subscription>();

		
		// Find all subscriptions on neighboring identifiers, but not on the
		// link. Those are the ones we are interested in.
		notOnLinkSet.addAll(link.getIdent1().getSubscriptions());
		notOnLinkSet.addAll(link.getIdent2().getSubscriptions());
		notOnLinkSet.removeAll(link.getSubscriptions());
				
		for (Subscription sub : notOnLinkSet) {
			if (m.matchesFilter(sub.getSearchRequest().getMatchLinksFilter())) {
				clearContainers(sub);
				ret.add(sub);
			}
		}
		return ret;
	}

	/**
	 * If a {@link SubscriptionObserver} calls getPollResultFor() the corresponding
	 * Publisher with the given sessionId is searched.
	 * 
	 * All of it's subscriptions are checked, to see if any of them has changes.
	 * If a subscription has changes either the search result, the update result,
	 * the delete result or the notify result has elements which might be of
	 * interest for the MAPC.
	 * 
	 * A {@link PollResult} is constructed which contains {@link SearchResult}
	 * objects for all the different result types and further for all
	 * subscriptions a client has.
	 * 
	 * thoughts:
	 * 	could this get too big?
	 * 
	 * @param sessionId
	 * @return
	 * @throws PollResultsTooBigException 
	 */
	public PollResult getPollResultFor(String sessionId) throws PollResultsTooBigException {
		Publisher pub = publisherRep.getPublisherBySessionId(sessionId);
		SubscriptionState subState = pub.getSubscriptionState();
		PollResultImpl ret = null;
		
		if (!subState.isNotified())
			throw new RuntimeException("getPollResultFor() but never notified");
		
		try {
			if (subState.isPollResultsTooBig()) {
				sLogger.debug(sName + ": PollResultsTooBig for "  + pub.getPublisherId());
				subState.resetPollResult();
				throw new PollResultsTooBigException("too big");
			}
			ret = subState.getPollResult();
			subState.resetPollResult();
			
			if (ret.isEmpty())
				throw new RuntimeException("empty pollResult for poll");

			// Remove subscriptions that lead to an error
			for (Subscription sub : subState.getSubscriptions())
				if (sub.exceededSize())
					subState.removeSubscription(sub);
			
			return ret;
		} finally {
			subState.unsetNotified();
		}
	}
	

	/**
	 * Set the {@link SubscriptionObserver} to the given one.
	 * 
	 * @param subObs
	 * @throws AlreadyObservedException if another {@link SubscriptionObserver} was
	 *  registered before
	 */
	public void setSubscriptionObserver(SubscriptionObserver subObs) throws AlreadyObservedException {
		if (mObserver != null)
			throw new AlreadyObservedException("Only one SubscriptionObserver is allowed!");
		
		mObserver = subObs;
	}

	/**
	 * Simple helper method which calls pollResultAvailable() on the observer
	 * with the sessionId of the given publisher.
	 * 
	 * @param pub
	 */
	private void notifyChangedSubscriptions(Publisher pub) {
		NullCheck.check(mObserver, "observer is null");
		
		SubscriptionState subState = pub.getSubscriptionState();
			
		// sanity check
		if (pub.getSessionId() == null)
			throw new RuntimeException("PROGRAMMING ERROR: "
					+ "Publisher " + pub.getPublisherId() + " has no session");
		
		if (!subState.isNotified()) {
			sLogger.trace(sName + ": " + pub.getPublisherId() + 
					" has new poll results");
			subState.setNotified();
			mObserver.pollResultAvailable(pub.getSessionId());
		}
	}
	
	
	/**
	 * Helper method to prepare the notification of the observer with all
	 * publishers that have new results available.
	 */
	private void doNotifications() {
		sLogger.debug(sName + ": Notify SubscriptionObserver about new results...");
		for (Publisher p : changedPublishers) {
			sLogger.trace(sName + ": "+ p.getPublisherId()+ " has changed subscriptions");
			notifyChangedSubscriptions(p);
		}
	}

	
	/**
	 * Go through all changed subscriptions and check whether one grew too big.
	 * If one did, mark it as such and remove the subscription from the graph.
	 * This way it won't be checked the next time, but will still exist in the
	 * publisher.
	 */
	private void markSubscriptionsGrewTooBig() {
		for (Subscription sub : changedSubscriptions) {
			Publisher pub = sub.getPublisherReference();
			SubscriptionState subState = pub.getSubscriptionState();
			SearchRequest sreq = sub.getSearchRequest();
			String name = sub.getName();
			
			PollResultImpl pollResult = subState.getPollResult();
			int size = pollResult.getByteCountFor(name);
			int maxSize = DataModelService.getServerConfiguration().getDefaultMaxSearchResultSize();
			
			if (sreq.maxSizeGiven())
				maxSize = sreq.getMaxResultSize();
			
			if (size > maxSize) {
				pollResult.removeResults(name);
				pollResult.addErrorResult(name);
				sub.setExceededSize();
				clearContainers(sub);
				sLogger.trace(sName + ": Subscription " + name + " of " + 
						pub.getPublisherId() + " grew too big");
			}
		}
	}

	/**
	 * If a publisher stores more than its set max-poll-result-size,
	 * he gets marked by us.
	 */
	private void markPublisherPollResultTooBig() {
		for (Publisher pub : changedPublishers) {
			SubscriptionState subState = pub.getSubscriptionState();
			PollResultImpl pr = subState.getPollResult();
			
			
			if (subState.isPollResultsTooBig()) {
				sLogger.trace(sName + ": " + pub.getPublisherId() + 
						" PollResult grew too big (was=" + pr.getByteCount() + 
						"bytes allowed=" + subState.getMaxPollSize() + " bytes)");
				
				// Remove all subscriptions from this publisher
				removeSubscriptions(pub);
				subState.setPollResultsTooBig();
			}
		}
	}

	/**
	 * Remove all active subscriptions of the given {@link Publisher} and
	 * further remove the subscriptions from the graph.
	 * 
	 * @param pub
	 */
	void removeSubscriptions(Publisher pub) {
		NullCheck.check(pub, "pub is null");
		SubscriptionState subState = pub.getSubscriptionState();
		
		for (Subscription sub : subState.getSubscriptions())
			clearContainers(sub);
	
		subState.clearSubscriptions();
		subState.unsetNotified();
		subState.resetPollResult();
	}
	
	/**
	 * Go through all {@link MetadataContainer} objects that this subscription
	 * is attached to and remove the reference to the given {@link Subscription}
	 * object.
	 * 
	 * @param sub
	 */
	void clearContainers(Subscription sub) {
		for (MetadataContainer mc : sub.getContainers()) {
			mc.removeSubscription(sub);
			sub.removeContainer(mc);
		}
		
		// Sanity
		if (sub.getContainers().size() != 0)
			throw new RuntimeException("Subscription" + sub.getName() + 
					" should not have any containers anymore");
	}

	/**
	 * Go through all {@link MetadataContainer} objects in the {@link SearchResult}
	 * and register the {@link Subscription} with them. Add the {@link MetadataContainer}
	 * objects to the subscription as well.
	 * 
	 * @param sub
	 * @param newRes
	 */
	private void registerContainers(Subscription sub, SearchResult newRes) {
		for (ResultItem ri : newRes.getResultItems()) {
			ri.getContainer().addSubscription(sub);
			sub.addContainer(ri.getContainer());
		}
	}

	private void cleanup() {
		changedSubscriptions.clear();
		changedPublishers.clear();
		mChangedMetadata.clear();
		mRerunSubs.clear();
	}

	
	/**
	 * FIXME: UNUSED?
	 * Rollback changes done in the graph.
	 * Every {@link Metadata} object which has state NEW has to be removed
	 * from the graph. Every {@link Metadata} object which has state DELETED
	 * needs to be set to state UNCHANGED.
	 */
	void rollbackChanges() {
		sLogger.debug(sName + ": Rollback changes made to metadata");
		
		for (Metadata m : mChangedMetadata) {
			if (m.isDeleted()) {
				m.setState(MetaState.UNCHANGED);
			} else if (m.isNew()) {
				metaController.removeReferencesTo(m);
			} else if (m.isNotify()) {
				metaController.removeReferencesTo(m);
			} else {
				throw new RuntimeException("Encountered unchanged Metadata during rollback. " +
						"Programming Error?");
			}
		}
	}
}
