/*
 * Project: irond
 * Package: src.de.fhhannover.inform.iron.mapserver.datamodel
 * File:    SubscriptionState.java
 *
 * Copyright (C) 2010-2011 Fachhochschule Hannover
 * Ricklinger Stadtweg 118, 30459 Hannover, Germany 
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.fhhannover.inform.iron.mapserver.datamodel;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import de.fhhannover.inform.iron.mapserver.IfmapConstStrings;
import de.fhhannover.inform.iron.mapserver.messages.PollResult;
import de.fhhannover.inform.iron.mapserver.utils.NullCheck;

/**
 * Class to encapsulate all state concerning subscriptions for a single
 * {@link Publisher}
 * 
 * @author aw
 */
class SubscriptionState {
	
	private boolean mNotified;
	private Integer  maxPollResultSize;
	private boolean mPollResGrewTooBig;
	
	private final Map<String, Subscription> mSubscriptions;
	private PollResultImpl mPollResult;
	
	SubscriptionState() {
		mSubscriptions = new HashMap<String, Subscription>();
		mPollResult = new PollResultImpl();
		mPollResGrewTooBig = false;
	}
	
	/**
	 * @return read-only copy of the {@link Subscription} instances
	 */
	List<Subscription> getSubscriptions() {
		return Collections.unmodifiableList(
				new ArrayList<Subscription>(mSubscriptions.values()));
	}
	
	/**
	 * Add a {@link Subscription} to this {@link SubscriptionState}.
	 * Goes crazy if the {@link Subscription} existed.
	 * 
	 * @param sub
	 */
	void addSubscription(Subscription sub) {
		NullCheck.check(sub, "sub is null");
		if (mSubscriptions.containsKey(sub.getName()))
			throw new RuntimeException("subscription " + sub.getName() + " exists!");
		
		mSubscriptions.put(sub.getName(), sub);
	}
	
	/**
	 * Removes the given {@link Subscription} from this {@link SubscriptionState}
	 * and cleans the {@link PollResult}.
	 * 
	 * @param sub
	 */
	void removeSubscription(Subscription sub) {
		NullCheck.check(sub, "sub is null");
		NullCheck.check(sub.getName(), "sub.getName() returns null");
		
		if (!mSubscriptions.containsKey(sub.getName()))
			throw new RuntimeException("Cannot remove nonexistent subscription"
					+ sub.getName());
		
		mSubscriptions.remove(sub.getName());
		mPollResult.removeResults(sub.getName());
	}

	/**
	 * @param name
	 * @return the subscription with name name or null
	 */
	public Subscription getSubscription(String name) {
		NullCheck.check(name, "name is null");
		return mSubscriptions.get(name);
	}

	/**
	 * Remove all {@link Subscription} instances and clean the {@link PollResult}.
	 */
	void clearSubscriptions() {
		for (Subscription sub : getSubscriptions())
			removeSubscription(sub);
	
		// Sanity
		if (mSubscriptions.size() != 0)
			throw new RuntimeException("All subscriptions should be gone");
		
		if (mPollResult.getResults().size() != 0)
			throw new RuntimeException("PollResult should be empty");
		
		if (mPollResult.getByteCount() != IfmapConstStrings.PRES_MIN_CNT)
			throw new RuntimeException("PollResult size should be zero");
	}
	
	Integer getMaxPollSize() {
		return maxPollResultSize;
	}

	void setMaxPollResultSize(Integer mprs) {
		maxPollResultSize = mprs;
	}
	
	void setNotified() {
		mNotified = true;
	}
	
	void unsetNotified() {
		mNotified = false;
	}
	
	boolean isNotified() {
		return mNotified;
	}

	void setPollResultsTooBig() {
		mPollResGrewTooBig = true;
	}
	
	void unsetPollResultsTooBig() {
		mPollResGrewTooBig = false;
	}
	
	boolean isPollResultsTooBig() {
		if (mPollResult.getByteCount() > getMaxPollSize())
			mPollResGrewTooBig = true;
		
		return (mPollResult.getByteCount() > getMaxPollSize())
				|| mPollResGrewTooBig;
	}
	
	void resetPollResult() {
		unsetPollResultsTooBig();
		mPollResult = new PollResultImpl();
	}
	
	PollResultImpl getPollResult() {
		return mPollResult;
	}

}
