/*
 * Project: irond
 * Package: src.de.fhhannover.inform.iron.mapserver.messages
 * File:    ResultItem.java
 *
 * Copyright (C) 2010-2011 Fachhochschule Hannover
 * Ricklinger Stadtweg 118, 30459 Hannover, Germany 
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.fhhannover.inform.iron.mapserver.messages;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import de.fhhannover.inform.iron.mapserver.datamodel.Filter;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.MetadataContainer;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.meta.Metadata;
import de.fhhannover.inform.iron.mapserver.utils.NullCheck;

/**
 * A {@link ResultItem} is a simple object encapsulating a {@link MetadataContainer}
 * and a list of {@link Metadata} objects.
 * 
 * @author aw
 *
 */
public class ResultItem {
	
	/**
	 * Represents the {@link MetadataContainer} in the {@link ResultItem}.
	 */
	private MetadataContainer mMetadataContainer;
	
	/**
	 * Represents the {@link Metadata} objects. Can be empty.
	 */
	private List<Metadata> mMetadata;

	public ResultItem(MetadataContainer mc) {
		NullCheck.check(mc, "mc not allowed to be null");
		mMetadataContainer = mc;
		mMetadata = new ArrayList<Metadata>();
	}
	
	public MetadataContainer getContainer() {
		return mMetadataContainer;
	}
	
	/**
	 * Add a {@link Metadata} object to this {@link ResultItem}. No copying.
	 * @param md
	 */
	public void addMetadata(Metadata md) {
		NullCheck.check(md, "md not allowed to be null");
		mMetadata.add(md);
	}
	
	/**
	 * @return a read-only list of the {@link Metadata} objects
	 */
	public List<Metadata> getMetdata() {
		return Collections.unmodifiableList(mMetadata);
	}

	/**
	 * Remove all {@link Metadata} objects from this {@link ResultItem} object
	 * if it does not match the given {@link Filter}.
	 * 
	 * @param resultFilter
	 */
	public void filter(Filter resultFilter) {
		final List<Metadata> toRemove = new ArrayList<Metadata>();
		for (Metadata m : mMetadata)
			if (!m.matchesFilter(resultFilter))
				toRemove.add(m);
		
		mMetadata.removeAll(toRemove);
	}

	@Override
	public String toString() {
		final StringBuilder sb = new StringBuilder();
		int i = 0;
		sb.append("ri{");
		sb.append(mMetadataContainer);
		sb.append(" #md=");
		sb.append(mMetadata.size());
		if (mMetadata.size() > 0) {
			sb.append(" (");
			for (Metadata m : mMetadata) {
				sb.append(m.getPrefixAndElement());
				i++;
				if (i != mMetadata.size())
					sb.append(", ");
			}
			sb.append(")");
		}
	
		sb.append("}");
		return sb.toString();
		
	}
}
