/*
 * Project: irond
 * Package: src.de.fhhannover.inform.iron.mapserver.messages
 * File:    SearchResult.java
 *
 * Copyright (C) 2010-2011 Fachhochschule Hannover
 * Ricklinger Stadtweg 118, 30459 Hannover, Germany 
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.fhhannover.inform.iron.mapserver.messages;

import java.util.List;

import de.fhhannover.inform.iron.mapserver.IfmapConstStrings;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.MetadataContainer;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.meta.Metadata;
import de.fhhannover.inform.iron.mapserver.utils.LengthCheck;
import de.fhhannover.inform.iron.mapserver.utils.NullCheck;

/**
 * Abstract superclass which contains abstract methods that need to be
 * implemented by subclasses and further some logic that can probably be
 * reused.
 * 
 * @author aw
 * @version 0.1
 *
 */

/*
 * created: long time ago
 * changes:
 * 	   05.10 aw - extracting a abstract class
 *  08.06.10 aw - changed LinkedList to ArrayList
 */
public abstract class SearchResult {
	
	/**
	 * The name of this {@link SearchResult}. May be null.
	 */
	private String mName;
	
	/**
	 * The type of this {@link SearchResult}.
	 */
	private SearchResultType mType;

	/**
	 * Constructor to create {@link SearchResult} instances.
	 * 
	 * @param name
	 * @param type
	 */
	public SearchResult(String name, SearchResultType type) {
		NullCheck.check(type, "type is null");
		
		if (name != null)
			LengthCheck.checkMinMax(name, 1, 20, "name is set, so length should" +
					" be > 0 && <= 20");
		
		if (name != null)
			mName = name;
		
		mType = type;
	}
	
	/**
	 * Constructor to create a {@link SearchResult} with name and type
	 * {@link SearchResultType#SEARCH}.
	 * 
	 * @param name
	 */
	public SearchResult(String name) {
		this(name, SearchResultType.SEARCH);
	}
	
	/**
	 * Constructor to create a {@link SearchResult} without a name and type
	 * {@link SearchResultType#SEARCH}.
	 */
	public SearchResult() {
		this(null, SearchResultType.SEARCH);
	}
	
	/**
	 * Add metadata objects to the given metadata container. A implementation
	 * of {@link SearchResult} is responsible to make deep copies of both,
	 * {@link MetadataContainer} objects and {@link Metadata} objects.
	 * 
	 * If the addressed {@link MetadataContainer} object was the last used
	 * {@link MetadataContainer} for a call to {@link #addMetadata(MetadataContainer, Metadata)}
	 * or {@link #addMetadataContainer(MetadataContainer)}, this entry can
	 * be reused, otherwise a new one needs to be created to fullfill the
	 * ordering requirement of metadata.
	 * 
	 * @param mc
	 * @param m
	 */
	public abstract void addMetadata(MetadataContainer mc, Metadata m);
	
	/**
	 * Only add a MetadataContainer. A implementation of {@link SearchResult}
	 * is responsible to make a deep copy of the {@link MetadataContainer} object.
	 * @param mc
	 */
	public abstract void addMetadataContainer(MetadataContainer mc);
	
	/**
	 * @return a unmodifieable of {@link MetadataContainer} instances contained
	 *         in this SearchResult. These {@link MetadataContainer} instances
	 *         represent the content of resultItem elements.
	 */
	public abstract List<ResultItem> getResultItems();
	
	/**
	 * @return list of all {@link MetadataContainer} instances contained in this
	 *         {@link SearchResult}.
	 *
	public List<MetadataContainer> getContainers() {
		List<MetadataContainer> ret = new ArrayList<MetadataContainer>();
		for (ResultItem ri : getResultItems())
			ret.add(ri.getMetadataContainer());
		
		return Collections.unmodifiableList(ret);
	}
	*/

	/**
	 * Returns the name of the corresponding subscription or
	 * null if no name was given.
	 * 
	 * @return
	 */
	public String getName() {
		return mName;
	}
	

	public boolean isEmpty() {
		return getResultItems().isEmpty();
	}
	
	/**
	 * @return the size of the elements, <b>including</b> the size for the
	 *         enclosing searchResult, updateResult, deleteResult or notifyResult
	 *         element.
	 */
	public int getByteCount() {
		return getSizeForEnclosingElement() + getSizeForResultItems();
	}
	
	public int getSizeForResultItems() {
		MetadataContainer mc = null;
		int ret = 0;
		for (ResultItem ri: getResultItems()) {
			mc = ri.getContainer();
			ret += IfmapConstStrings.RITEM_MIN_CNT;
			ret += mc.getByteCount();
			ret += getSizeForMetadata(ri);
		}
		return ret;
	}

	public final SearchResultType getType() {
		return mType;
	}

	/**
	 * Helper to add a number of {@link ResultItem} objects to this
	 * {@link SearchResult}.
	 * 
	 * @param resultItems
	 */
	public void addResultItems(List<ResultItem> resultItems) {
		NullCheck.check(resultItems, "resultItems is null");
		
		for (ResultItem ri : resultItems)
			addResultItem(ri);
	}

	/**
	 * Helper to add a single {@link ResultItem} object to this
	 * {@link SearchResult}.
	 * 
	 * @param ri
	 */
	public void addResultItem(ResultItem ri) {
		NullCheck.check(ri, "ri is null");
		if (ri.getMetdata().size() == 0) {
			addMetadataContainer(ri.getContainer());
		} else {
			addMetadata(ri.getContainer(), ri.getMetdata());
		}
	}

	/**
	 * Helper to add multiple metadata to this {@link SearchResult}, the
	 * contract is as for {@link #addMetadata(MetadataContainer, Metadata)}.
	 * 
	 * @param container
	 * @param metadata
	 */
	public void addMetadata(MetadataContainer container, List<Metadata> metadata) {
		NullCheck.check(container, "container is null");
		NullCheck.check(metadata, "metadata is null");
		
		for (Metadata m : metadata)
			addMetadata(container, m);
	}
	
	private int getSizeForMetadata(ResultItem ri) {
		int ret = 0;
		List<Metadata> mdlist = ri.getMetdata();
		if (mdlist.size() > 0) {
			ret += IfmapConstStrings.MLIST_MIN_CNT;
			for (Metadata m : mdlist) {
				ret += m.getByteCount();
			}
		}
		return ret;
	}

	private int getSizeForEnclosingElement() {
		final int ret;
		
		// They are all the same, but make it explicit by switching
		switch (getType()) {
			case SEARCH:
				ret = IfmapConstStrings.SRES_MIN_CNT;
				break;
			case UPDATE:
				ret = IfmapConstStrings.URES_MIN_CNT;
				break;
			case DELETE:
				ret = IfmapConstStrings.DRES_MIN_CNT;
				break;
			case NOTIFY:
				ret = IfmapConstStrings.NRES_MIN_CNT;
				break;
			default:
				throw new RuntimeException("Unknown searchResult type");
		}
		
		return ret;
	}

	public String toString() {
		// FIXME: Implement this in a nice way.
		return "SerachResult.toString() FIXME";
		/*
		StringBuffer sb = new StringBuffer();
		sb.append("Identifiers:\n");
		for (Identifier i : getIdentifiers()) {
			sb.append("\t" + i.toString() + "\n");
			for (Metadata m : getMetadataFor(i)) {
				sb.append("\t\t" + m.getType().getTypeString() + "\n");
			}
		}
		sb.append("Links:\n");
		for (Link l : getLinks()) {
			sb.append("\t" + l.toString() + "\n");
			for (Metadata m : getMetadataFor(l)) {
				sb.append("\t\t" + m.getType().getTypeString() + "\n");
			}
		}
		return sb.toString();
		*/
	}
}
